unit Unit1; 

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Menus;

type

  { TForm1 }

  TForm1 = class(TForm)
    Button1: TButton;
    OpenDialog1: TOpenDialog;
    OpenMenuItem: TMenuItem;
    PopupMenu1: TPopupMenu;
    SaveDialog1: TSaveDialog;
    SaveMenuItem: TMenuItem;
    SumEdit: TEdit;
    LeftOverEdit: TEdit;
    Label4: TLabel;
    Label5: TLabel;
    WantedEdit: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    ResultListBox: TListBox;
    InputMemo: TMemo;
    procedure Button1Click(Sender: TObject);
    procedure OpenMenuItemClick(Sender: TObject);
    procedure SaveMenuItemClick(Sender: TObject);
  private
    fInNums: array of cardinal;
    fWanted, fBestMatchLen, fBestCombination: cardinal;
    procedure Calculate;
    function ShowResults: Boolean;
  public

  end; 

var
  Form1: TForm1; 

implementation

{$R *.lfm}

{ TForm1 }

procedure TForm1.Calculate;
var
  TotalLen: cardinal;

  function CheckBestMatch(aCombination: cardinal): Boolean;
  // Return True if the Length matches exactly with fWanted.
  begin
    Result := False;
    if TotalLen < fBestMatchLen then begin
      fBestMatchLen := TotalLen;
      fBestCombination := aCombination;
      // If perfect match is found, rest will be skipped.
      Result := TotalLen = fWanted;
    end;
  end;

var
  CmbCount, Cmb, BitCounter, i: Cardinal;
  DoChecks: Boolean;
begin
  CmbCount := 1 shl Length(fInNums);     // 2 ^ count combinations.
  Cmb := 0;
  // Loop through all the combinations, 2 at a time.
  while Cmb < CmbCount do begin
    DoChecks := True;
    TotalLen := 0;
    BitCounter := 2;  // = 1 shl 1
    // Loop through input numbers.
    // Add the number if the corresponding bit in combination is set.
    for i := 1 to Length(fInNums)-1 do begin // Skip first item now, added later.
      if (Cmb and BitCounter) > 0 then begin
        Inc(TotalLen, fInNums[i]);
        if TotalLen >= fBestMatchLen then begin // Went over, no need to check more.
          DoChecks := False;
          Break;
        end;
      end;
      BitCounter := BitCounter shl 1;  // Move the mask bit step by step to left.
    end;
    if DoChecks then begin
      if TotalLen >= fWanted then begin
        if CheckBestMatch(Cmb) then Break;   // Check how well the length matches.
      end
      else begin
        Inc(TotalLen, fInNums[0]);     // Now add the first num (LSB bit, Cmb+1).
        if TotalLen >= fWanted then begin
          if CheckBestMatch(Cmb+1) then Break; // Check the Cmb+1.
        end;
      end;
    end;
    Inc(Cmb, 2);    // Did 2 combinations on the same round.
  end;
end;

function TForm1.ShowResults: Boolean;
var
  BitCounter, i: cardinal;
begin
  Result := True;
  BitCounter := 1;
  ResultListBox.Items.Clear;
  for i := 0 to Length(fInNums)-1 do begin
    if (fBestCombination and BitCounter) > 0 then
      with ResultListBox.Items do
        Add(Format('%s (%s.)', [IntToStr(fInNums[i]),IntToStr(i+1)]));
    BitCounter := BitCounter shl 1;
  end;
  // Sum and leftover.
  SumEdit.Text := IntToStr(fBestMatchLen);
  LeftOverEdit.Text := IntToStr(fBestMatchLen-fWanted);
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  i: Integer;
  s: string;
  RealCount: cardinal;
begin
  fBestMatchLen := MaxInt;            // This is bigger than any real length.
  fBestCombination := 0;
  // User defined numbers to an array:
  RealCount := 0;
  SetLength(fInNums,InputMemo.Lines.Count);
  for i := 0 to InputMemo.Lines.Count-1 do begin
    s := InputMemo.Lines[i];
    if s <> '' then begin
      fInNums[i] := StrToInt(s);
      Inc(RealCount);
    end;
  end;
  if RealCount > 0 then begin
    if RealCount > SizeOf(Cardinal) * 8 then begin
      ShowMessage(Format('Numeroita saa olla korkeintaan %d. Nyt niitä on %d',
                         [i, RealCount]));
      Exit;
    end;
    SetLength(fInNums,RealCount);     // Make sure the array has only valid items.
    fWanted := StrToInt(WantedEdit.Text);
    Screen.Cursor:=crHourGlass;
    try
      Calculate;                      // Data is in right place, calculate ...
    finally
      Screen.Cursor:=crDefault;
    end;
    if fBestMatchLen <> MaxInt then
      ShowResults
    else
      ShowMessage('Ei löytynyt.');
  end
  else
    ShowMessage('Ei numeroita.');
end;

procedure TForm1.OpenMenuItemClick(Sender: TObject);
begin
  if OpenDialog1.Execute then
    InputMemo.Lines.LoadFromFile(OpenDialog1.FileName);
end;

procedure TForm1.SaveMenuItemClick(Sender: TObject);
begin
  if SaveDialog1.Execute then
    InputMemo.Lines.SaveToFile(SaveDialog1.FileName);
end;


end.

